'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
__doc__ = """
The ``pm4py.ocel`` module contains the object-centric process mining features offered in ``pm4py``
"""

from typing import List, Dict, Collection, Any, Optional, Set, Tuple

import pandas as pd

from pm4py.objects.ocel.obj import OCEL
from pm4py.util import constants, pandas_utils
import sys
import random


def ocel_get_object_types(ocel: OCEL) -> List[str]:
    """
    Gets the list of object types contained in the object-centric event log
    (e.g., ["order", "item", "delivery"]).

    :param ocel: object-centric event log
    :rtype: ``List[str]``

    .. code-block:: python3

        import pm4py

        object_types = pm4py.ocel_get_object_types(ocel)
    """
    return pandas_utils.format_unique(ocel.objects[ocel.object_type_column].unique())


def ocel_get_attribute_names(ocel: OCEL) -> List[str]:
    """
    Gets the list of attributes at the event and the object level of an object-centric event log
    (e.g. ["cost", "amount", "name"])

    :param ocel: object-centric event log
    :rtype: ``List[str]``

    .. code-block:: python3

        import pm4py

        attribute_names = pm4py.ocel_get_attribute_names(ocel)
    """
    from pm4py.objects.ocel.util import attributes_names
    return attributes_names.get_attribute_names(ocel)


def ocel_flattening(ocel: OCEL, object_type: str) -> pd.DataFrame:
    """
    Flattens the object-centric event log to a traditional event log with the choice of an object type.
    In the flattened log, the objects of a given object type are the cases, and each case
    contains the set of events related to the object.

    :param ocel: object-centric event log
    :param object_type: object type
    :rtype: ``pd.DataFrame``

    .. code-block:: python3

        import pm4py

        event_log = pm4py.ocel_flattening(ocel, 'items')
    """
    from pm4py.objects.ocel.util import flattening
    return flattening.flatten(ocel, object_type)


def ocel_object_type_activities(ocel: OCEL) -> Dict[str, Collection[str]]:
    """
    Gets the set of activities performed for each object type

    :param ocel: object-centric event log
    :rtype: ``Dict[str, Collection[str]]``

    .. code-block:: python3

        import pm4py

        ot_activities = pm4py.ocel_object_type_activities(ocel)
    """
    from pm4py.statistics.ocel import ot_activities

    return ot_activities.get_object_type_activities(ocel)


def ocel_objects_ot_count(ocel: OCEL) -> Dict[str, Dict[str, int]]:
    """
    Counts for each event the number of related objects per type

    :param ocel: object-centric event log
    :rtype: ``Dict[str, Dict[str, int]]``

    .. code-block:: python3

        import pm4py

        objects_ot_count = pm4py.ocel_objects_ot_count(ocel)
    """
    from pm4py.statistics.ocel import objects_ot_count

    return objects_ot_count.get_objects_ot_count(ocel)


def ocel_temporal_summary(ocel: OCEL) -> pd.DataFrame:
    """
    Returns the ``temporal summary'' from an object-centric event log.
    The temporal summary aggregates all the events performed in the same timestamp,
    and reports the list of activities and the involved objects.

    :param ocel: object-centric event log
    :rtype: ``pd.DataFrame``

    .. code-block:: python3

        import pm4py

        temporal_summary = pm4py.ocel_temporal_summary(ocel)
    """
    gdf = ocel.relations.groupby(ocel.event_timestamp)
    act_comb = gdf[ocel.event_activity].agg(list).to_frame()
    obj_comb = gdf[ocel.object_id_column].agg(list).to_frame()
    temporal_summary = act_comb.join(obj_comb).reset_index()
    return temporal_summary


def ocel_objects_summary(ocel: OCEL) -> pd.DataFrame:
    """
    Gets the objects summary of an object-centric event log

    :param ocel: object-centric event log
    :rtype: ``pd.DataFrame``

    .. code-block:: python3

        import pm4py

        objects_summary = pm4py.ocel_objects_summary(ocel)
    """
    gdf = ocel.relations.groupby(ocel.object_id_column)
    act_comb = gdf[ocel.event_activity].agg(list).to_frame().rename(columns={ocel.event_activity: "activities_lifecycle"})
    lif_start_tim = gdf[ocel.event_timestamp].min().to_frame().rename(columns={ocel.event_timestamp: "lifecycle_start"})
    lif_end_tim = gdf[ocel.event_timestamp].max().to_frame().rename(columns={ocel.event_timestamp: "lifecycle_end"})
    objects_summary = act_comb.join(lif_start_tim)
    objects_summary = objects_summary.join(lif_end_tim)
    objects_summary = objects_summary.reset_index()
    objects_summary["lifecycle_duration"] = pandas_utils.get_total_seconds(objects_summary["lifecycle_end"] - objects_summary["lifecycle_start"])
    ev_rel_obj = ocel.relations.groupby(ocel.event_id_column)[ocel.object_id_column].agg(list).to_dict()
    objects_ids = pandas_utils.format_unique(ocel.objects[ocel.object_id_column].unique())
    graph = {o: set() for o in objects_ids}
    for ev in ev_rel_obj:
        rel_obj = ev_rel_obj[ev]
        for o1 in rel_obj:
            for o2 in rel_obj:
                if o1 != o2:
                    graph[o1].add(o2)
    objects_summary["interacting_objects"] = objects_summary[ocel.object_id_column].map(graph)
    return objects_summary


def ocel_objects_interactions_summary(ocel: OCEL) -> pd.DataFrame:
    """
    Gets the objects interactions summary of an object-centric event log.
    The objects interactions summary has a row for every combination (event, related object, other related object).
    Properties such as the activity of the event, and the object types of the two related objects, are included.

    :param ocel: object-centric event log
    :rtype: ``OCEL``

    .. code-block:: python3

        import pm4py

        interactions_summary = pm4py.ocel_objects_interactions_summary(ocel)
    """
    obj_types = ocel.objects.groupby(ocel.object_id_column)[ocel.object_type_column].first().to_dict()
    eve_activities = ocel.events.groupby(ocel.event_id_column)[ocel.event_activity].first().to_dict()
    ev_rel_obj = ocel.relations.groupby(ocel.event_id_column)[ocel.object_id_column].agg(list).to_dict()
    stream = []
    for ev in ev_rel_obj:
        rel_obj = ev_rel_obj[ev]
        for o1 in rel_obj:
            for o2 in rel_obj:
               if o1 != o2:
                   stream.append({ocel.event_id_column: ev, ocel.event_activity: eve_activities[ev],
                                  ocel.object_id_column: o1, ocel.object_type_column: obj_types[o1],
                                  ocel.object_id_column+"_2": o2, ocel.object_type_column+"_2": obj_types[o2]})

    return pandas_utils.instantiate_dataframe(stream)


def discover_ocdfg(ocel: OCEL, business_hours=False, business_hour_slots=constants.DEFAULT_BUSINESS_HOUR_SLOTS) -> Dict[str, Any]:
    """
    Discovers an OC-DFG from an object-centric event log.

    Object-centric directly-follows multigraphs are a composition of directly-follows graphs for the single object type, which can be annotated with different metrics considering the entities of an object-centric event log (i.e., events, unique objects, total objects).

    Reference paper:
    Berti, Alessandro, and Wil van der Aalst. "Extracting multiple viewpoint models from relational databases." Data-Driven Process Discovery and Analysis. Springer, Cham, 2018. 24-51.

    :param ocel: object-centric event log
    :param business_hours: boolean value that enables the usage of the business hours
    :param business_hour_slots: work schedule of the company, provided as a list of tuples where each tuple represents one time slot of business hours. One slot i.e. one tuple consists of one start and one end time given in seconds since week start, e.g. [(7 * 60 * 60, 17 * 60 * 60), ((24 + 7) * 60 * 60, (24 + 12) * 60 * 60), ((24 + 13) * 60 * 60, (24 + 17) * 60 * 60),] meaning that business hours are Mondays 07:00 - 17:00 and Tuesdays 07:00 - 12:00 and 13:00 - 17:00

    :rtype: ``Dict[str, Any]``

    .. code-block:: python3

        import pm4py

        ocdfg = pm4py.discover_ocdfg(ocel)
    """
    parameters = {}
    parameters["business_hours"] = business_hours
    parameters["business_hour_slots"] = business_hour_slots
    from pm4py.algo.discovery.ocel.ocdfg import algorithm as ocdfg_discovery
    return ocdfg_discovery.apply(ocel, parameters=parameters)


def discover_oc_petri_net(ocel: OCEL, inductive_miner_variant: str = "im", diagnostics_with_tbr: bool = False) -> Dict[str, Any]:
    """
    Discovers an object-centric Petri net from the provided object-centric event log.

    Reference paper: van der Aalst, Wil MP, and Alessandro Berti. "Discovering object-centric Petri nets." Fundamenta informaticae 175.1-4 (2020): 1-40.

    :param ocel: object-centric event log
    :param inductive_miner_variant: specify the variant of the inductive miner to be used
                            ("im" for traditional; "imd" for the faster inductive miner directly-follows)
    :param diagnostics_with_tbr: (boolean) enables the computation of some diagnostics using token-based replay
    :rtype: ``Dict[str, Any]``

    .. code-block:: python3

        import pm4py

        ocpn = pm4py.discover_oc_petri_net(ocel)
    """
    from pm4py.algo.discovery.ocel.ocpn import algorithm as ocpn_discovery
    parameters = {}
    parameters["inductive_miner_variant"] = inductive_miner_variant
    parameters["diagnostics_with_token_based_replay"] = diagnostics_with_tbr

    return ocpn_discovery.apply(ocel, parameters=parameters)


def discover_objects_graph(ocel: OCEL, graph_type: str = "object_interaction") -> Set[Tuple[str, str]]:
    """
    Discovers an object graph from the provided object-centric event log

    :param ocel: object-centric event log
    :param graph_type: type of graph to consider (object_interaction, object_descendants, object_inheritance, object_cobirth, object_codeath)
    :rtype: ``Dict[str, Any]``

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        obj_graph = pm4py.ocel_discover_objects_graph(ocel, graph_type='object_interaction')
    """
    if graph_type == "object_interaction":
        from pm4py.algo.transformation.ocel.graphs import object_interaction_graph
        return object_interaction_graph.apply(ocel)
    elif graph_type == "object_descendants":
        from pm4py.algo.transformation.ocel.graphs import object_descendants_graph
        return object_descendants_graph.apply(ocel)
    elif graph_type == "object_inheritance":
        from pm4py.algo.transformation.ocel.graphs import object_inheritance_graph
        return object_inheritance_graph.apply(ocel)
    elif graph_type == "object_cobirth":
        from pm4py.algo.transformation.ocel.graphs import object_cobirth_graph
        return object_cobirth_graph.apply(ocel)
    elif graph_type == "object_codeath":
        from pm4py.algo.transformation.ocel.graphs import object_codeath_graph
        return object_codeath_graph.apply(ocel)


def ocel_o2o_enrichment(ocel: OCEL, included_graphs: Optional[Collection[str]] = None) -> OCEL:
    """
    Inserts the information inferred from the graph computations (pm4py.discover_objects_graph)
    in the list of O2O relations of the OCEL.

    :param ocel: object-centric event log
    :param included_graphs: types of graphs to include, provided as list/set of strings (object_interaction_graph, object_descendants_graph, object_inheritance_graph, object_cobirth_graph, object_codeath_graph)
    :rtype: ``OCEL``


    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        ocel = pm4py.ocel_o2o_enrichment(ocel)
        print(ocel.o2o)
    """
    from pm4py.algo.transformation.ocel.graphs import ocel20_computation
    return ocel20_computation.apply(ocel, parameters={"included_graphs": included_graphs})


def ocel_e2o_lifecycle_enrichment(ocel: OCEL) -> OCEL:
    """
    Inserts lifecycle-based information (when an object is created/terminated or other types of relations)
    in the list of E2O relations of the OCEL

    :param ocel: object-centric event log
    :rtype: ``OCEL``

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        ocel = pm4py.ocel_e2o_lifecycle_enrichment(ocel)
        print(ocel.relations)
    """
    from pm4py.objects.ocel.util import e2o_qualification
    ocel = e2o_qualification.apply(ocel, "termination")
    ocel = e2o_qualification.apply(ocel, "creation")
    ocel = e2o_qualification.apply(ocel, "other")
    return ocel


def sample_ocel_objects(ocel: OCEL, num_objects: int) -> OCEL:
    """
    Given an object-centric event log, returns a sampled event log with a subset of the objects
    that is chosen in a random way.
    Only the events related to at least one of these objects are filtered from the event log.
    As a note, the relationships between the different objects are probably going to be ruined by
    this sampling.

    :param ocel: Object-centric event log
    :param num_objects: Number of objects of the object-centric event log
    :rtype: ``OCEL``

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        sampled_ocel = pm4py.sample_ocel_objects(ocel, 50) # keeps only 50 random objects
    """
    from pm4py.objects.ocel.util import sampling
    return sampling.sample_ocel_objects(ocel, parameters={"num_entities": num_objects})


def sample_ocel_connected_components(ocel: OCEL, connected_components: int = 1,
                                     max_num_events_per_cc: int = sys.maxsize,
                                     max_num_objects_per_cc: int = sys.maxsize,
                                     max_num_e2o_relations_per_cc: int = sys.maxsize) -> OCEL:
    """
    Given an object-centric event log, returns a sampled event log with a subset of the executions.
    The number of considered connected components need to be specified by the user.

    Paper:
    Adams, Jan Niklas, et al. "Defining cases and variants for object-centric event data." 2022 4th International Conference on Process Mining (ICPM). IEEE, 2022.

    :param ocel: Object-centric event log
    :param connected_components: Number of connected components to pick from the OCEL
    :param max_num_events_per_cc: maximum number of events allowed per connected component (default: sys.maxsize)
    :param max_num_objects_per_cc: maximum number of events allowed per connected component (default: sys.maxsize)
    :param max_num_e2o_relations_per_cc: maximum number of event-to-object relationships allowed per connected component (default: sys.maxsize)
    :rtype: ``OCEL``

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        sampled_ocel = pm4py.sample_ocel_connected_components(ocel, 5) # keeps only 5 connected components
    """
    from pm4py.algo.transformation.ocel.split_ocel import algorithm
    ocel_splits = algorithm.apply(ocel, variant=algorithm.Variants.CONNECTED_COMPONENTS)
    events = None
    objects = None
    relations = None
    ocel_splits = [x for x in ocel_splits if
                   len(x.events) <= max_num_events_per_cc and len(x.objects) <= max_num_objects_per_cc and len(
                       x.relations) <= max_num_e2o_relations_per_cc]

    if len(ocel_splits) > 0:
        ocel_splits = random.sample(ocel_splits, min(connected_components, len(ocel_splits)))

    for cc in ocel_splits:
        if events is None:
            events = cc.events
            objects = cc.objects
            relations = cc.relations
        else:
            events = pandas_utils.concat([events, cc.events])
            objects = pandas_utils.concat([objects, cc.objects])
            relations = pandas_utils.concat([relations, cc.relations])

    return OCEL(events, objects, relations)


def ocel_drop_duplicates(ocel: OCEL) -> OCEL:
    """
    Drop relations between events and objects happening at the same time,
    with the same activity, to the same object identifier.
    This ends up cleaning the OCEL from duplicate events.

    :param ocel: object-centric event log
    :rtype: ``OCEL``

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        ocel = pm4py.ocel_drop_duplicates(ocel)

    """
    from pm4py.objects.ocel.util import filtering_utils
    ocel.relations = ocel.relations.drop_duplicates(
        subset=[ocel.event_activity, ocel.event_timestamp, ocel.object_id_column])
    ocel = filtering_utils.propagate_relations_filtering(ocel)
    return ocel


def ocel_merge_duplicates(ocel: OCEL, have_common_object: Optional[bool]=False) -> OCEL:
    """
    Merge events in the OCEL that happen with the same activity at the same timestamp

    :param ocel: object-centric event log
    :param have_common_object: impose the additional merge condition that the two events should happen at the same
                                timestamp.
    :rtype: ``OCEL``

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        ocel = pm4py.ocel_merge_duplicates(ocel)
    """
    import copy
    import uuid
    relations = copy.copy(ocel.relations)
    if have_common_object:
        relations["@@groupn"] = relations.groupby([ocel.object_id_column, ocel.event_activity, ocel.event_timestamp]).ngroup()
    else:
        relations["@@groupn"] = relations.groupby([ocel.event_activity, ocel.event_timestamp]).ngroup()

    group_size = relations["@@groupn"].value_counts().to_dict()
    relations["@@groupsize"] = relations["@@groupn"].map(group_size)
    relations = relations.sort_values(["@@groupsize", "@@groupn"], ascending=False)
    val_corr = {x: str(uuid.uuid4()) for x in pandas_utils.format_unique(relations["@@groupn"].unique())}
    relations = relations.groupby(ocel.event_id_column).first()["@@groupn"].to_dict()
    relations = {x: val_corr[y] for x, y in relations.items()}

    ocel.events[ocel.event_id_column] = ocel.events[ocel.event_id_column].map(relations)
    ocel.relations[ocel.event_id_column] = ocel.relations[ocel.event_id_column].map(relations)

    ocel.events = ocel.events.drop_duplicates(subset=[ocel.event_id_column])
    ocel.relations = ocel.relations.drop_duplicates(subset=[ocel.event_id_column, ocel.object_id_column])

    return ocel



def ocel_sort_by_additional_column(ocel: OCEL, additional_column: str, primary_column: str = "ocel:timestamp") -> OCEL:
    """
    Sorts the OCEL not only based on the timestamp column and the index,
    but using an additional sorting column that further determines the order of
    the events happening at the same timestamp.

    :param ocel: object-centric event log
    :param additional_column: additional column to use for the sorting
    :param primary_column: primary column to be used for the sorting (default: ocel:timestamp)
    :rtype: ``OCEL``

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        ocel = pm4py.ocel_sort_by_additional_column(ocel, 'ordering')

    """
    ocel.events = pandas_utils.insert_index(ocel.events, "@@index", reset_index=False, copy_dataframe=False)
    ocel.events = ocel.events.sort_values([primary_column, additional_column, "@@index"])
    del ocel.events["@@index"]
    ocel.events = ocel.events.reset_index(drop=True)
    return ocel


def ocel_add_index_based_timedelta(ocel: OCEL) -> OCEL:
    """
    Adds a small time-delta to the timestamp column based on the current index of the event.
    This ensures the correct ordering of the events in any object-centric process mining
    solution.

    :param ocel: object-centric event log
    :rtype: ``OCEL``

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        ocel = pm4py.ocel_add_index_based_timedelta(ocel)

    """
    from datetime import timedelta
    eids = ocel.events[ocel.event_id_column].to_numpy().tolist()
    eids = {eids[i]: timedelta(milliseconds=i) for i in range(len(eids))}
    ocel.events["@@timedelta"] = ocel.events[ocel.event_id_column].map(eids)
    ocel.relations["@@timedelta"] = ocel.relations[ocel.event_id_column].map(eids)
    ocel.events[ocel.event_timestamp] = ocel.events[ocel.event_timestamp] + ocel.events["@@timedelta"]
    ocel.relations[ocel.event_timestamp] = ocel.relations[ocel.event_timestamp] + ocel.relations["@@timedelta"]
    del ocel.events["@@timedelta"]
    del ocel.relations["@@timedelta"]
    return ocel


def cluster_equivalent_ocel(ocel: OCEL, object_type: str, max_objs: int = sys.maxsize) -> Dict[str, Collection[OCEL]]:
    """
    Perform a clustering of the object-centric event log, based on the 'executions' of
    a single object type. Equivalent 'executions' are grouped in the output dictionary.

    :param ocel: object-centric event log
    :param object_type: reference object type
    :param max_objs: maximum number of objects (of the given object type)
    :rtype: ``Dict[str, Collection[OCEL]]``

    .. code-block:: python3

        import pm4py

        ocel = pm4py.read_ocel('trial.ocel')
        clusters = pm4py.cluster_equivalent_ocel(ocel, "order")
    """
    from pm4py.algo.transformation.ocel.split_ocel import algorithm as split_ocel_algorithm
    from pm4py.objects.ocel.util import rename_objs_ot_tim_lex
    from pm4py.algo.transformation.ocel.description import algorithm as ocel_description
    lst_ocels = split_ocel_algorithm.apply(ocel, variant=split_ocel_algorithm.Variants.ANCESTORS_DESCENDANTS, parameters={"object_type": object_type, "max_objs": max_objs})
    ret = {}
    for index, oc in enumerate(lst_ocels):
        oc_ren = rename_objs_ot_tim_lex.apply(oc)
        descr = ocel_description.apply(oc_ren, parameters={"include_timestamps": False})
        if descr not in ret:
            ret[descr] = []
        ret[descr].append(oc)
    return ret



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from collections import Counter
from enum import Enum
from typing import Optional, List, Dict, Any, Tuple, Union

import numpy as np

from pm4py.objects.log.obj import EventLog
from pm4py.util import exec_utils
from pm4py.objects.conversion.log import converter


class Parameters(Enum):
    STR_EVENT_ATTRIBUTES = "str_ev_attr"
    NUM_EVENT_ATTRIBUTES = "num_ev_attr"
    FEATURE_NAMES = "feature_names"
    MIN_NUM_DIFF_STR_VALUES = "min_num_diff_str_values"
    MAX_NUM_DIFF_STR_VALUES = "max_num_diff_str_values"


def extract_all_ev_features_names_from_log(log: EventLog, str_ev_attr: List[str], num_ev_attr: List[str],
                                           parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> List[str]:
    """
    Extracts the feature names from an event log.

    Parameters
    ---------------
    log
        Event log
    str_ev_attr
        (if provided) list of string event attributes to consider in extracting the feature names
    num_ev_attr
        (if provided) list of integer event attributes to consider in extracting the feature names
    parameters
        Parameters, including:
            - MIN_NUM_DIFF_STR_VALUES => minimum number of distinct values to include an attribute as feature(s)
            - MAX_NUM_DIFF_STR_VALUES => maximum number of distinct values to include an attribute as feature(s)

    Returns
    ----------------
    feature_names
        List of feature names
    """
    if parameters is None:
        parameters = {}

    min_num_diff_str_values = exec_utils.get_param_value(Parameters.MIN_NUM_DIFF_STR_VALUES, parameters, 2)
    max_num_diff_str_values = exec_utils.get_param_value(Parameters.MAX_NUM_DIFF_STR_VALUES, parameters, 500)

    str_features = {}
    num_features = Counter()
    count_events = 0
    for trace in log:
        for event in trace:
            count_events += 1
            for attr_name in event:
                attr_value = event[attr_name]
                if isinstance(attr_value, str) and (str_ev_attr is None or attr_name in str_ev_attr):
                    if attr_name not in str_features:
                        str_features[attr_name] = set()
                    str_features[attr_name].add("event:" + attr_name + "@" + attr_value)
                elif isinstance(attr_value, int) or isinstance(attr_value, float):
                    if num_ev_attr is None or attr_name in num_ev_attr:
                        num_features["event:" + attr_name] += 1
    num_features = list({x for x, y in num_features.items() if y == count_events})
    str_features = list({z for x, y in str_features.items() for z in y if
                         min_num_diff_str_values <= len(y) <= max_num_diff_str_values})

    feature_names = str_features + num_features
    feature_names = sorted(feature_names)

    return feature_names


def extract_features(log: EventLog, feature_names: List[str], parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> Tuple[
    Any, List[str]]:
    """
    Extracts the matrix of the features from an event log

    Parameters
    ---------------
    log
        Event log
    feature_names
        Features to consider (in the given order)

    Returns
    -------------
    data
        Data to provide for decision tree learning
    feature_names
        Names of the features, in order
    """
    v1 = max(len(trace) for trace in log)
    v2 = len(feature_names)
    data = np.zeros((len(log), v1, v2), dtype=np.float32)

    for i1, trace in enumerate(log):
        for i2, event in enumerate(trace):
            str_features = set()
            num_features = {}
            for attr_name in event:
                attr_value = event[attr_name]
                if isinstance(attr_value, str):
                    str_features.add("event:" + attr_name + "@" + attr_value)
                elif isinstance(attr_value, int) or isinstance(attr_value, float):
                    num_features["event:" + attr_name] = float(attr_value)
            for attr in str_features:
                if attr in feature_names:
                    data[i1, i2, feature_names.index(attr)] = 1.0
            for attr in num_features:
                if attr in feature_names:
                    data[i1, i2, feature_names.index(attr)] = float(num_features[attr])

    return data, feature_names


def apply(log: EventLog, parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> Tuple[Any, List[str]]:
    """
    Extracts all the features for the traces of an event log (each trace becomes a vector of vectors, where each
    event has its own vector)

    Parameters
    -----------------
    log
        Event log
    parameters
        Parameters of the algorithm, including:
            - STR_EVENT_ATTRIBUTES => string event attributes to consider in the features extraction
            - NUM_EVENT_ATTRIBUTES => numeric event attributes to consider in the features extraction
            - FEATURE_NAMES => features to consider (in the given order)

    Returns
    -------------
    data
        Data to provide for decision tree learning
    feature_names
        Names of the features, in order
    """
    if parameters is None:
        parameters = {}

    str_ev_attr = exec_utils.get_param_value(Parameters.STR_EVENT_ATTRIBUTES, parameters, None)
    num_ev_attr = exec_utils.get_param_value(Parameters.NUM_EVENT_ATTRIBUTES, parameters, None)
    feature_names = exec_utils.get_param_value(Parameters.FEATURE_NAMES, parameters, None)

    log = converter.apply(log, variant=converter.Variants.TO_EVENT_LOG, parameters=parameters)

    if feature_names is None:
        feature_names = extract_all_ev_features_names_from_log(log, str_ev_attr, num_ev_attr, parameters=parameters)

    return extract_features(log, feature_names, parameters=parameters)



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''

from pm4py.objects.petri_net.obj import PetriNet
from typing import Optional, Dict, Any


def apply(petri_net: PetriNet, parameters: Optional[Dict[Any, Any]] = None) -> float:
    """
    Computes the extended Cardoso metric as described in the paper:

    "Complexity Metrics for Workflow Nets"
    Lassen, Kristian Bisgaard, and Wil MP van der Aalst

    Parameters
    -------------
    petri_net
        Petri net

    Returns
    -------------
    ext_cardoso_metric
        Extended Cardoso metric
    """
    if parameters is None:
        parameters = {}

    ext_card = 0

    for place in petri_net.places:
        targets = set()
        for out_arc in place.out_arcs:
            for out_arc2 in out_arc.target.out_arcs:
                targets.add(out_arc2.target)
        ext_card += len(targets)

    return ext_card



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''

from pm4py.algo.evaluation.simplicity.variants import arc_degree, extended_cardoso, extended_cyclomatic
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any
from pm4py.objects.petri_net.obj import PetriNet


class Variants(Enum):
    SIMPLICITY_ARC_DEGREE = arc_degree
    EXTENDED_CARDOSO = extended_cardoso
    EXTENDED_CYCLOMATIC = extended_cyclomatic


SIMPLICITY_ARC_DEGREE = Variants.SIMPLICITY_ARC_DEGREE
EXTENDED_CARDOSO = Variants.EXTENDED_CARDOSO
EXTENDED_CYCLOMATIC = Variants.EXTENDED_CYCLOMATIC

VERSIONS = {SIMPLICITY_ARC_DEGREE, EXTENDED_CARDOSO, EXTENDED_CYCLOMATIC}


def apply(petri_net: PetriNet, parameters: Optional[Dict[Any, Any]] = None, variant=SIMPLICITY_ARC_DEGREE) -> float:
    return exec_utils.get_variant(variant).apply(petri_net, parameters=parameters)



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from collections import Counter
from enum import Enum

from pm4py.algo.discovery.log_skeleton import trace_skel
from pm4py.objects.log.util import xes
from pm4py.util import exec_utils
from pm4py.util import variants_util, pandas_utils
from pm4py.util.constants import PARAMETER_CONSTANT_ACTIVITY_KEY, PARAMETER_CONSTANT_CASEID_KEY, CASE_CONCEPT_NAME
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog
import pandas as pd


class Parameters(Enum):
    # parameter for the noise threshold
    NOISE_THRESHOLD = "noise_threshold"
    # considered constraints in conformance checking among: equivalence, always_after, always_before, never_together, directly_follows, activ_freq
    CONSIDERED_CONSTRAINTS = "considered_constraints"
    # default choice for conformance checking
    DEFAULT_CONSIDERED_CONSTRAINTS = ["equivalence", "always_after", "always_before", "never_together",
                                      "directly_follows", "activ_freq"]
    CASE_ID_KEY = PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    PARAMETER_VARIANT_DELIMITER = "variant_delimiter"


NOISE_THRESHOLD = Parameters.NOISE_THRESHOLD
CONSIDERED_CONSTRAINTS = Parameters.CONSIDERED_CONSTRAINTS
DEFAULT_CONSIDERED_CONSTRAINTS = Parameters.DEFAULT_CONSIDERED_CONSTRAINTS
ACTIVITY_KEY = Parameters.ACTIVITY_KEY
PARAMETER_VARIANT_DELIMITER = Parameters.PARAMETER_VARIANT_DELIMITER


class Outputs(Enum):
    EQUIVALENCE = "equivalence"
    ALWAYS_AFTER = "always_after"
    ALWAYS_BEFORE = "always_before"
    NEVER_TOGETHER = "never_together"
    DIRECTLY_FOLLOWS = "directly_follows"
    ACTIV_FREQ = "activ_freq"


def equivalence(logs_traces, all_activs, noise_threshold=0):
    """
    Gets the equivalence relations given the traces of the log

    Parameters
    -------------
    logs_traces
        Traces of the log
    all_activs
        All the activities
    noise_threshold
        Noise threshold

    Returns
    --------------
    rel
        List of relations in the log
    """
    ret0 = Counter()
    for trace in logs_traces:
        rs = Counter(trace_skel.equivalence(list(trace)))
        for k in rs:
            rs[k] = rs[k] * logs_traces[trace]
        ret0 += rs
    ret = set(x for x, y in ret0.items() if y >= all_activs[x[0]] * (1.0 - noise_threshold))
    return ret


def always_after(logs_traces, all_activs, noise_threshold=0):
    """
    Gets the always-after relations given the traces of the log

    Parameters
    -------------
    logs_traces
        Traces of the log
    all_activs
        All the activities
    noise_threshold
        Noise threshold

    Returns
    --------------
    rel
        List of relations in the log
    """
    ret0 = Counter()
    for trace in logs_traces:
        rs = Counter(trace_skel.after(list(trace)))
        for k in rs:
            rs[k] = rs[k] * logs_traces[trace]
        ret0 += rs
    ret = set(x for x, y in ret0.items() if y >= all_activs[x[0]] * (1.0 - noise_threshold))
    return ret


def always_before(logs_traces, all_activs, noise_threshold=0):
    """
    Gets the always-before relations given the traces of the log

    Parameters
    -------------
    logs_traces
        Traces of the log
    all_activs
        All the activities
    noise_threshold
        Noise threshold

    Returns
    --------------
    rel
        List of relations in the log
    """
    ret0 = Counter()
    for trace in logs_traces:
        rs = Counter(trace_skel.before(list(trace)))
        for k in rs:
            rs[k] = rs[k] * logs_traces[trace]
        ret0 += rs
    ret = set(x for x, y in ret0.items() if y >= all_activs[x[0]] * (1.0 - noise_threshold))
    return ret


def never_together(logs_traces, all_activs, len_log, noise_threshold=0):
    """
    Gets the never-together relations given the traces of the log

    Parameters
    -------------
    logs_traces
        Traces of the log
    all_activs
        All the activities
    len_log
        Length of the log
    noise_threshold
        Noise threshold

    Returns
    --------------
    rel
        List of relations in the log
    """
    all_combos = set((x, y) for x in all_activs for y in all_activs if x != y)
    ret0 = Counter()
    for k in all_combos:
        ret0[k] = all_activs[k[0]]
    for trace in logs_traces:
        rs = Counter(trace_skel.combos(list(trace)))
        for k in rs:
            rs[k] = rs[k] * logs_traces[trace]
        ret0 -= rs
    ret = set(x for x, y in ret0.items() if y >= all_activs[x[0]] * (1.0 - noise_threshold))
    return ret


def directly_follows(logs_traces, all_activs, noise_threshold=0):
    """
    Gets the allowed directly-follows relations given the traces of the log

    Parameters
    -------------
    logs_traces
        Traces of the log
    all_activs
        All the activities
    noise_threshold
        Noise threshold

    Returns
    --------------
    rel
        List of relations in the log
    """
    ret0 = Counter()
    for trace in logs_traces:
        rs = Counter(trace_skel.directly_follows(list(trace)))
        for k in rs:
            rs[k] = rs[k] * logs_traces[trace]
        ret0 += rs
    ret = set(x for x, y in ret0.items() if y >= all_activs[x[0]] * (1.0 - noise_threshold))
    return ret


def activ_freq(logs_traces, all_activs, len_log, noise_threshold=0):
    """
    Gets the allowed activities frequencies given the traces of the log

    Parameters
    -------------
    logs_traces
        Traces of the log
    all_activs
        All the activities
    len_log
        Length of the log
    noise_threshold
        Noise threshold

    Returns
    --------------
    rel
        List of relations in the log
    """
    ret0 = {}
    ret = {}
    for trace in logs_traces:
        rs = trace_skel.activ_freq(trace)
        for act in all_activs:
            if act not in rs:
                rs[act] = 0
        for act in rs:
            if act not in ret0:
                ret0[act] = Counter()
            ret0[act][rs[act]] += logs_traces[trace]
    for act in ret0:
        ret0[act] = sorted(list((x, y) for x, y in ret0[act].items()), key=lambda x: x[1], reverse=True)
        added = 0
        i = 0
        while i < len(ret0[act]):
            added += ret0[act][i][1]
            if added >= (1.0 - noise_threshold) * len_log:
                ret0[act] = ret0[act][:min(i + 1, len(ret0[act]))]
            i = i + 1
        ret[act] = set(x[0] for x in ret0[act])
    return ret


def apply(log: Union[EventLog, pd.DataFrame], parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> Dict[str, Any]:
    """
    Discover a log skeleton from an event log

    Parameters
    -------------
    log
        Event log
    parameters
        Parameters of the algorithm, including:
            - the activity key (Parameters.ACTIVITY_KEY)
            - the noise threshold (Parameters.NOISE_THRESHOLD)

    Returns
    -------------
    model
        Log skeleton model
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes.DEFAULT_NAME_KEY)
    noise_threshold = exec_utils.get_param_value(Parameters.NOISE_THRESHOLD, parameters, 0.0)

    if type(log) is EventLog:
        logs_traces = Counter([tuple(y[activity_key] for y in x) for x in log])
        all_activs = Counter(list(y[activity_key] for x in log for y in x))
    elif pandas_utils.check_is_pandas_dataframe(log):
        case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME)
        all_activs = log[activity_key].value_counts().to_dict()
        logs_traces = Counter([tuple(x) for x in log.groupby(case_id_key)[activity_key].agg(list).to_dict().values()])

    ret = {}
    ret[Outputs.EQUIVALENCE.value] = equivalence(logs_traces, all_activs, noise_threshold=noise_threshold)
    ret[Outputs.ALWAYS_AFTER.value] = always_after(logs_traces, all_activs, noise_threshold=noise_threshold)
    ret[Outputs.ALWAYS_BEFORE.value] = always_before(logs_traces, all_activs, noise_threshold=noise_threshold)
    ret[Outputs.NEVER_TOGETHER.value] = never_together(logs_traces, all_activs, len(log),
                                                       noise_threshold=noise_threshold)
    ret[Outputs.DIRECTLY_FOLLOWS.value] = directly_follows(logs_traces, all_activs, noise_threshold=noise_threshold)
    ret[Outputs.ACTIV_FREQ.value] = activ_freq(logs_traces, all_activs, len(log), noise_threshold=noise_threshold)

    return ret


def apply_from_variants_list(var_list, parameters=None):
    """
    Discovers the log skeleton from the variants list

    Parameters
    ---------------
    var_list
        Variants list
    parameters
        Parameters

    Returns
    ---------------
    model
        Log skeleton model
    """
    if parameters is None:
        parameters = {}

    log = EventLog()
    for cv in var_list:
        v = cv[0]
        trace = variants_util.variant_to_trace(v, parameters=parameters)
        log.append(trace)

    return apply(log, parameters=parameters)


def prepare_encode(log_skeleton):
    """
    Prepares the log skeleton for encoding

    Parameters
    --------------
    log_skeleton
        Log skeleton

    Returns
    --------------
    log_skeleton
        Log skeleton (with lists instead of sets)
    """
    log_skeleton[Outputs.EQUIVALENCE.value] = list(log_skeleton[Outputs.EQUIVALENCE.value])
    log_skeleton[Outputs.ALWAYS_AFTER.value] = list(log_skeleton[Outputs.ALWAYS_AFTER.value])
    log_skeleton[Outputs.ALWAYS_BEFORE.value] = list(log_skeleton[Outputs.ALWAYS_BEFORE.value])
    log_skeleton[Outputs.NEVER_TOGETHER.value] = list(log_skeleton[Outputs.NEVER_TOGETHER.value])
    log_skeleton[Outputs.DIRECTLY_FOLLOWS.value] = list(log_skeleton[Outputs.DIRECTLY_FOLLOWS.value])
    for act in log_skeleton[Outputs.ACTIV_FREQ.value]:
        log_skeleton[Outputs.ACTIV_FREQ.value][act] = list(log_skeleton[Outputs.ACTIV_FREQ.value][act])
    return log_skeleton



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
import copy
import sys
from typing import Dict, Tuple, Optional, List

from pm4py.objects.process_tree.utils import generic as ptu
from pm4py.objects.process_tree.obj import Operator
from pm4py.objects.process_tree.obj import ProcessTree

ProcessTreeState = Dict[Tuple[int, ProcessTree], ProcessTree.OperatorState]


def get_initial_state(tree: ProcessTree) -> ProcessTreeState:
    state = dict()
    state[(id(tree), tree)] = ProcessTree.OperatorState.FUTURE
    path = [(tree, ProcessTree.OperatorState.FUTURE)]
    e_path, state = enable_vertex(tree, state)
    path.extend(e_path)
    return state


def transform_tree(tree: ProcessTree, state_type: ProcessTree.OperatorState, state: ProcessTreeState) -> Tuple[
    List[Tuple[ProcessTree, ProcessTree.OperatorState]], ProcessTreeState]:
    state = copy.copy(state)
    path = list()
    if (id(tree), tree) not in state or state[(id(tree), tree)] != state_type:
        state[(id(tree), tree)] = state_type
        path = [(tree, state_type)]
    for c in tree.children:
        e_path, state = transform_tree(c, state_type, state)
        path.extend(e_path)
    return path, state


def can_enable(tree: ProcessTree, state: ProcessTreeState) -> bool:
    if ptu.is_in_state(tree, ProcessTree.OperatorState.FUTURE, state):
        if ptu.is_root(tree):
            return True
        if ptu.is_in_state(tree.parent, ProcessTree.OperatorState.OPEN, state):
            if ptu.is_any_operator_of(tree.parent, [Operator.PARALLEL, Operator.OR]):
                return True
            elif ptu.is_operator(tree.parent, Operator.XOR):
                return frozenset(map(lambda child: state[(id(child), child)], tree.parent.children)) == {
                    ProcessTree.OperatorState.FUTURE}
            elif ptu.is_operator(tree.parent, Operator.SEQUENCE):
                return True if tree.parent.children.index(tree) == 0 else ptu.is_in_state(
                    tree.parent.children[tree.parent.children.index(tree) - 1],
                    ProcessTree.OperatorState.CLOSED, state)
            elif ptu.is_operator(tree.parent, Operator.LOOP):
                return frozenset(map(lambda child: state[(id(child), child)], tree.parent.children)) == {
                    ProcessTree.OperatorState.FUTURE, ProcessTree.OperatorState.CLOSED}
    return False


def can_open(tree: ProcessTree, state: ProcessTreeState) -> bool:
    return ptu.is_in_state(tree, ProcessTree.OperatorState.ENABLED, state)


def can_close(tree: ProcessTree, state: ProcessTreeState) -> bool:
    if ptu.is_leaf(tree):
        return ptu.is_in_state(tree, ProcessTree.OperatorState.OPEN, state)
    elif ptu.is_any_operator_of(tree, [Operator.SEQUENCE, Operator.PARALLEL, Operator.XOR]):
        return frozenset(map(lambda child: state[(id(child), child)], tree.children)) == {
            ProcessTree.OperatorState.CLOSED}
    elif ptu.is_any_operator_of(tree, [Operator.OR]):
        return frozenset(map(lambda child: state[(id(child), child)], tree.children)) == {
            ProcessTree.OperatorState.CLOSED, ProcessTree.OperatorState.FUTURE}
    elif ptu.is_operator(tree, Operator.LOOP):
        return ptu.is_in_state(tree.children[0], ProcessTree.OperatorState.CLOSED, state) and ptu.is_in_state(
            tree.children[1], ProcessTree.OperatorState.FUTURE, state)


def close_vertex(tree: ProcessTree, state: ProcessTreeState) -> Tuple[Optional[
                                                                          List[Tuple[
                                                                              ProcessTree, ProcessTree.OperatorState]]],
                                                                      Optional[ProcessTreeState]]:
    if can_close(tree, state):
        current_state = state[(id(tree), tree)]
        path = list()
        state = copy.copy(state)
        for c in tree.children:
            if not ptu.is_in_state(c, ProcessTree.OperatorState.CLOSED, state):
                e_path, state = transform_tree(c, ProcessTree.OperatorState.CLOSED, state)
                path.extend(e_path)
        state[(id(tree), tree)] = ProcessTree.OperatorState.CLOSED
        path.append((tree, ProcessTree.OperatorState.CLOSED))
        # if tree is a redo, then we will always need to execute the do part of the surrounding loop
        if ptu.is_operator(tree.parent, Operator.LOOP) and id(tree) == id(tree.parent.children[
            1]) and current_state == ProcessTree.OperatorState.OPEN:
            e_path, state = enable_vertex(tree.parent.children[0], state)
            path.extend(e_path)
    else:
        state, path = None, None
    return path, state


def enable_vertex(tree: ProcessTree, state: ProcessTreeState) -> Tuple[Optional[
                                                                           List[Tuple[
                                                                               ProcessTree, ProcessTree.OperatorState]]],
                                                                       Optional[ProcessTreeState]]:
    if state[(id(tree), tree)] == ProcessTree.OperatorState.ENABLED:
        return list(), state
    if can_enable(tree, state):
        state = copy.copy(state)
        path = list()
        state[(id(tree), tree)] = ProcessTree.OperatorState.ENABLED
        path.append((tree, ProcessTree.OperatorState.ENABLED))
        if ptu.is_operator(tree.parent, Operator.LOOP):
            if id(tree) == id(tree.parent.children[0]):
                e_path, state = transform_tree(tree.parent.children[1], ProcessTree.OperatorState.FUTURE, state)
                path.extend(e_path)
            if id(tree) == id(tree.parent.children[1]):
                e_path, state = transform_tree(tree.parent.children[0], ProcessTree.OperatorState.FUTURE, state)
                path.extend(e_path)
        if ptu.is_operator(tree.parent, Operator.XOR):
            for c in tree.parent.children:
                if id(c) != id(tree):
                    e_path, state = transform_tree(c, ProcessTree.OperatorState.CLOSED, state)
                    path.extend(e_path)
        for c in tree.children:
            e_path, state = transform_tree(c, ProcessTree.OperatorState.FUTURE, state)
            path.extend(e_path)
        return path, state
    else:
        return None, None


def open_vertex(tree: ProcessTree, state: ProcessTreeState) -> Tuple[Optional[
                                                                         List[Tuple[
                                                                             ProcessTree, ProcessTree.OperatorState]]],
                                                                     Optional[ProcessTreeState]]:
    if can_open(tree, state):
        state = copy.copy(state)
        path = list()
        state[(id(tree), tree)] = ProcessTree.OperatorState.OPEN
        path.append((tree, ProcessTree.OperatorState.OPEN))
        if ptu.is_any_operator_of(tree, [Operator.XOR, Operator.OR, Operator.PARALLEL]):
            for c in tree.children:
                state[(id(c), c)] = ProcessTree.OperatorState.FUTURE
                path.append((c, ProcessTree.OperatorState.FUTURE))
        elif ptu.is_any_operator_of(tree, [Operator.SEQUENCE, Operator.LOOP]):
            state[(id(tree.children[0]), tree.children[0])] = ProcessTree.OperatorState.ENABLED
            path.append((tree.children[0], ProcessTree.OperatorState.ENABLED))
            for c in tree.children[1:]:
                state[(id(c), c)] = ProcessTree.OperatorState.FUTURE
                path.append((c, ProcessTree.OperatorState.FUTURE))
        return path, state
    else:
        return None, None


def shortest_path_to_open(tree: ProcessTree, state: ProcessTreeState) -> Tuple[
    List[Tuple[ProcessTree, ProcessTree.OperatorState]], ProcessTreeState]:
    if ptu.is_in_state(tree, ProcessTree.OperatorState.OPEN, state):
        return list(), state
    fast_path, fast_state = open_vertex(tree, state)
    if fast_state is not None:
        return fast_path, fast_state
    path, state = shortest_path_to_enable(tree, state)
    if path is not None:
        e_path, state = open_vertex(tree, state)
        path.extend(e_path)
    return path, state


def shortest_path_to_close(tree: ProcessTree, state: ProcessTreeState) -> Tuple[
    List[Tuple[ProcessTree, ProcessTree.OperatorState]], ProcessTreeState]:
    if ptu.is_in_state(tree, ProcessTree.OperatorState.CLOSED, state):
        return list(), state
    fast_path, fast_state = close_vertex(tree, state)
    if fast_state is not None:
        return fast_path, fast_state
    path, state = shortest_path_to_open(tree, state)
    if ptu.is_leaf(tree):
        e_path, state = close_vertex(tree, state)
        path.extend(e_path)
        return path, state
    elif ptu.is_any_operator_of(tree, [Operator.SEQUENCE, Operator.PARALLEL]):
        for c in tree.children:
            e_path, state = shortest_path_to_close(c, state)
            path.extend(e_path)
    elif ptu.is_operator(tree, Operator.LOOP):
        if state[(id(tree.children[0]), tree.children[0])] in {ProcessTree.OperatorState.ENABLED,
                                                               ProcessTree.OperatorState.OPEN}:
            e_path, state = shortest_path_to_close(tree.children[0], state)
            path.extend(e_path)
        elif state[(id(tree.children[1]), tree.children[1])] in {ProcessTree.OperatorState.ENABLED,
                                                                 ProcessTree.OperatorState.OPEN}:
            e_path, state = shortest_path_to_close(tree.children[1], state)
            path.extend(e_path)
            e_path, state = shortest_path_to_open(tree.children[0], state)
            path.extend(e_path)
            e_path, state = shortest_path_to_close(tree.children[0], state)
            path.extend(e_path)
    elif tree.operator in {Operator.XOR, Operator.OR}:
        busy = False
        for c in tree.children:
            if state[(id(c), c)] in {ProcessTree.OperatorState.ENABLED, ProcessTree.OperatorState.OPEN}:
                e_path, state = shortest_path_to_close(c, state)
                path.extend(e_path)
                busy = True
        if not busy:
            cur_path, cur_state, cur_path_costs = list(), copy.copy(state), sys.maxsize
            for c in tree.children:
                if state[(id(c), c)] != ProcessTree.OperatorState.CLOSED:
                    candidate_p, candidate_s = shortest_path_to_close(c, state)
                    candidate_costs = len(list(filter(lambda t: t[0].operator is None and t[0].label is not None and t[
                        1] == ProcessTree.OperatorState.OPEN, candidate_p)))
                    if candidate_costs < cur_path_costs:
                        cur_path, cur_state, cur_path_costs = candidate_p, candidate_s, candidate_costs
            path.extend(cur_path)
            state = cur_state
    e_path, state = close_vertex(tree, state)
    path.extend(e_path)
    return path, state


def shortest_path_to_enable(tree: ProcessTree, state: ProcessTreeState) -> Tuple[
    List[Tuple[ProcessTree, ProcessTree.OperatorState]], ProcessTreeState]:
    if state[(id(tree), tree)] == ProcessTree.OperatorState.ENABLED:
        return list(), state
    fast_path, fast_state = enable_vertex(tree, state)
    if fast_state is not None:
        return fast_path, fast_state
    if state[(id(tree), tree)] == ProcessTree.OperatorState.FUTURE:
        path, state = shortest_path_to_open(tree.parent, state)
        if tree.parent.operator in {Operator.XOR, Operator.PARALLEL, Operator.OR}:
            e_path, state = enable_vertex(tree, state)
            if state is not None:  # choice if another choice has already been taken!
                path.extend(e_path)
        if tree.parent.operator == Operator.SEQUENCE:
            for i, c in enumerate(tree.parent.children):
                if id(c) == id(tree):
                    if i > 0:
                        e_path, state = enable_vertex(tree, state)
                        path.extend(e_path)
                    break
                else:
                    e_path, state = shortest_path_to_close(c, state)
                    path.extend(e_path)
        elif tree.parent.operator == Operator.LOOP:
            if id(tree) == id(tree.parent.children[0]):
                if state[(id(tree.parent.children[1]), tree.parent.children[1])] not in {
                    ProcessTree.OperatorState.FUTURE,
                    ProcessTree.OperatorState.CLOSED}:
                    e_path, state = shortest_path_to_close(tree.parent.children[1], state)
                    path.extend(e_path)
                e_path, state = enable_vertex(tree, state)
                path.extend(e_path)
            else:
                e_path, state = shortest_path_to_close(tree.parent.children[0], state)
                path.extend(e_path)
                e_path, state = enable_vertex(tree, state)
                path.extend(e_path)
        return path, state
    else:
        path, state = shortest_path_to_close(tree, state)
        parent = tree.parent
        while parent is not None:
            if parent.operator == Operator.LOOP and state[(id(parent), parent)] == ProcessTree.OperatorState.OPEN:
                break
            parent = parent.parent
        if parent is not None and parent.operator == Operator.LOOP:
            if state[(id(parent.children[0]), parent.children[0])] == ProcessTree.OperatorState.OPEN:
                e_path, state = shortest_path_to_close(parent.children[0], state)
                path.extend(e_path)
                e_path, state = shortest_path_to_enable(parent.children[1], state)
                path.extend(e_path)
            elif state[(id(parent.children[1]), parent.children[1])] == ProcessTree.OperatorState.OPEN:
                e_path, state = shortest_path_to_close(parent.children[1], state)
                path.extend(e_path)
                e_path, state = shortest_path_to_enable(parent.children[0], state)
                path.extend(e_path)
            elif state[(id(parent.children[0]), parent.children[0])] == ProcessTree.OperatorState.FUTURE:
                e_path, state = shortest_path_to_enable(parent.children[0], state)
                path.extend(e_path)
            elif state[(id(parent.children[1]), parent.children[1])] == ProcessTree.OperatorState.FUTURE:
                e_path, state = shortest_path_to_enable(parent.children[1], state)
                path.extend(e_path)
            e_path, state = shortest_path_to_enable(tree, state)
            path.extend(e_path)
            return path, state
        else:
            return None, None



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''

from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
from enum import Enum
from pm4py.objects.ocel import constants
from pm4py.util import exec_utils, pandas_utils
from copy import deepcopy
import pandas as pd


class Parameters(Enum):
    EVENT_ID = constants.PARAM_EVENT_ID
    EVENT_ACTIVITY = constants.PARAM_EVENT_ACTIVITY
    EVENT_TIMESTAMP = constants.PARAM_EVENT_TIMESTAMP
    OBJECT_ID = constants.PARAM_OBJECT_ID
    OBJECT_TYPE = constants.PARAM_OBJECT_TYPE
    INTERNAL_INDEX = constants.PARAM_INTERNAL_INDEX


def apply(ocel: OCEL, param: str, parameters: Optional[Dict[Any, Any]] = None):
    """
    Transforms an event attribute to an object type.

    Parameters
    ---------------
    ocel
        Object-centric event log
    param
        Event attribute that should be moved at the event level
    parameters
        Parameters of the algorithm, including:
        - Parameters.EVENT_ID => the event identifier column
        - Parameters.EVENT_ACTIVITY => the event activity column
        - Parameters.EVENT_TIMESTAMP => the event timestamp column
        - Parameters.OBJECT_ID => the object identifier column
        - Parameters.OBJECT_TYPE => the object type column
        - Parameters.INTERNAL_INDEX => the internal index

    Returns
    --------------
    new_ocel
        OCEL in which the attribute has been moved to the object type level.
    """
    if parameters is None:
        parameters = {}

    event_id = exec_utils.get_param_value(Parameters.EVENT_ID, parameters, ocel.event_id_column)
    event_activity = exec_utils.get_param_value(Parameters.EVENT_ACTIVITY, parameters, ocel.event_activity)
    event_timestamp = exec_utils.get_param_value(Parameters.EVENT_TIMESTAMP, parameters, ocel.event_timestamp)
    object_id = exec_utils.get_param_value(Parameters.OBJECT_ID, parameters, ocel.object_id_column)
    object_type = exec_utils.get_param_value(Parameters.OBJECT_TYPE, parameters, ocel.object_type_column)
    internal_index = exec_utils.get_param_value(Parameters.INTERNAL_INDEX, parameters, constants.DEFAULT_INTERNAL_INDEX)

    new_ocel = deepcopy(ocel)
    new_ocel.events[param] = new_ocel.events[param].astype(str)
    ev_param = new_ocel.events.dropna(subset=[param])[[event_id, event_activity, event_timestamp, param]]
    vals = pandas_utils.format_unique(ev_param[param].unique())
    ev_param = ev_param.rename(columns={param: object_id})
    ev_param[object_type] = param

    new_objects_df = pandas_utils.instantiate_dataframe([{object_id: v, object_type: param} for v in vals])

    new_ocel.objects = pandas_utils.concat([new_ocel.objects, new_objects_df])
    new_ocel.relations = pandas_utils.concat([new_ocel.relations, ev_param])

    new_ocel.relations = pandas_utils.insert_index(new_ocel.relations, internal_index, reset_index=False, copy_dataframe=False)

    new_ocel.relations = new_ocel.relations.sort_values([event_timestamp, internal_index])

    del new_ocel.events[param]

    return new_ocel



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from pm4py.objects.log.obj import EventLog


def insert_event_index_as_event_attribute(stream, event_index_attr_name="@@eventindex"):
    """
    Insert the current event index as event attribute

    Parameters
    -----------
    stream
        Stream
    event_index_attr_name
        Attribute name given to the event index
    """

    if not type(stream) is EventLog:
        for i in range(0, len(stream._list)):
            stream._list[i][event_index_attr_name] = i + 1

    return stream


def insert_trace_index_as_event_attribute(log, trace_index_attr_name="@@traceindex"):
    """
    Inserts the current trace index as event attribute
    (overrides previous values if needed)

    Parameters
    -----------
    log
        Log
    trace_index_attr_name
        Attribute name given to the trace index
    """
    for i in range(len(log._list)):
        for j in range(len(log._list[i])):
            log._list[i][j][trace_index_attr_name] = i + 1

    return log



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from scipy.spatial.distance import squareform
import numpy as np
from pm4py.algo.clustering.trace_attribute_driven.variants import act_dist_calc
from pm4py.algo.clustering.trace_attribute_driven.variants import suc_dist_calc
from pm4py.algo.clustering.trace_attribute_driven.leven_dist import leven_dist_calc
from pm4py.algo.clustering.trace_attribute_driven.dfg import dfg_dist


def dfg_dis(loglist, percent, alpha):
    size = len(loglist)
    dist_mat = np.zeros((size, size))

    for i in range(0, size - 1):
        for j in range(i + 1, size):
            (dist_act, dist_dfg) = dfg_dist.dfg_dist_calc(loglist[i], loglist[j])
            dist_mat[i][j] = dist_act * alpha + dist_dfg * (1 - alpha)
            dist_mat[j][i] = dist_mat[i][j]
    y = squareform(dist_mat)
    return y


def eval_avg_variant(loglist, percent, alpha):
    size = len(loglist)
    dist_mat = np.zeros((size, size))

    for i in range(0, size - 1):
        for j in range(i + 1, size):
            dist_act = act_dist_calc.act_sim_percent_avg(loglist[i], loglist[j], percent, percent)
            dist_suc = suc_dist_calc.suc_sim_percent_avg(loglist[i], loglist[j], percent, percent)
            dist_mat[i][j] = dist_act * alpha + dist_suc * (1 - alpha)
            dist_mat[j][i] = dist_mat[i][j]
    y = squareform(dist_mat)

    return y


def eval_DMM_variant(loglist, percent, alpha):
    size = len(loglist)
    dist_mat = np.zeros((size, size))

    for i in range(0, size - 1):
        for j in range(i + 1, size):
            dist_act = act_dist_calc.act_sim_percent(loglist[i], loglist[j], percent, percent)
            dist_suc = suc_dist_calc.suc_sim_percent(loglist[i], loglist[j], percent, percent)
            dist_mat[i][j] = dist_act * alpha + dist_suc * (1 - alpha)
            dist_mat[j][i] = dist_mat[i][j]
    y = squareform(dist_mat)
    return y


def eval_avg_leven(loglist, percent, alpha):
    size = len(loglist)
    dist_mat = np.zeros((size, size))

    for i in range(0, size - 1):
        for j in range(i + 1, size):
            dist_mat[i][j] = leven_dist_calc.leven_dist_avg(loglist[i], loglist[j], percent, percent)
            dist_mat[j][i] = dist_mat[i][j]
    y = squareform(dist_mat)
    return y


def eval_DMM_leven(loglist, percent, alpha):
    size = len(loglist)
    dist_mat = np.zeros((size, size))

    for i in range(0, size - 1):
        for j in range(i + 1, size):
            dist_mat[i][j] = leven_dist_calc.leven_dist(loglist[i], loglist[j], percent, percent)
            dist_mat[j][i] = dist_mat[i][j]
    y = squareform(dist_mat)
    return y



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from typing import List, Optional, Tuple, TypeVar, Dict, Any

from pm4py.algo.discovery.inductive.cuts.abc import Cut
from pm4py.algo.discovery.inductive.cuts.concurrency import ConcurrencyCutUVCL, ConcurrencyCutDFG
from pm4py.algo.discovery.inductive.cuts.loop import LoopCutUVCL, LoopCutDFG
from pm4py.algo.discovery.inductive.cuts.sequence import StrictSequenceCutUVCL, StrictSequenceCutDFG, SequenceCutUVCL, SequenceCutDFG
from pm4py.algo.discovery.inductive.cuts.xor import ExclusiveChoiceCutUVCL, ExclusiveChoiceCutDFG
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructure, IMDataStructureUVCL, IMDataStructureDFG
from pm4py.algo.discovery.inductive.variants.instances import IMInstance
from pm4py.objects.process_tree.obj import ProcessTree
from pm4py.util import exec_utils
from enum import Enum


T = TypeVar('T', bound=IMDataStructure)
S = TypeVar('S', bound=Cut)


class Parameters(Enum):
    DISABLE_STRICT_SEQUENCE_CUT = "disable_strict_sequence_cut"


class CutFactory:

    @classmethod
    def get_cuts(cls, obj: T, inst: IMInstance, parameters: Optional[Dict[str, Any]] = None) -> List[S]:
        if parameters is None:
            parameters = {}

        disable_strict_sequence_cut = exec_utils.get_param_value(Parameters.DISABLE_STRICT_SEQUENCE_CUT, parameters, False)

        if inst is IMInstance.IM or inst is IMInstance.IMf:
            if type(obj) is IMDataStructureUVCL:
                sequence_cut = StrictSequenceCutUVCL
                if disable_strict_sequence_cut:
                    sequence_cut = SequenceCutUVCL
                return [ExclusiveChoiceCutUVCL, sequence_cut, ConcurrencyCutUVCL, LoopCutUVCL]
        if inst is IMInstance.IMd:
            if type(obj) is IMDataStructureDFG:
                sequence_cut = StrictSequenceCutDFG
                if disable_strict_sequence_cut:
                    sequence_cut = SequenceCutDFG
                return [ExclusiveChoiceCutDFG, sequence_cut, ConcurrencyCutDFG, LoopCutDFG]
        return list()

    @classmethod
    def find_cut(cls, obj: IMDataStructure, inst: IMInstance, parameters: Optional[Dict[str, Any]] = None) -> Optional[Tuple[ProcessTree, List[T]]]:
        for c in CutFactory.get_cuts(obj, inst, parameters):
            r = c.apply(obj, parameters)
            if r is not None:
                return r
        return None



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
import random
import string

from pm4py.objects.process_tree.obj import ProcessTree
from pm4py.objects.process_tree.obj import Operator
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union


class Parameters(Enum):
    REC_DEPTH = "rec_depth"
    MIN_REC_DEPTH = "min_rec_depth"
    MAX_REC_DEPTH = "max_rec_depth"
    PROB_LEAF = "prob_leaf"


def generate_random_string(N):
    """
    Generate a random string

    Parameters
    -------------
    N
        length of the string

    Returns
    -------------
    random_string
        Random string
    """
    return ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(N))


def get_random_operator():
    """
    Gets a random operator

    Returns
    ------------
    operator
        Operator
    """
    r = random.random()
    if r < 0.25:
        return Operator.SEQUENCE
    elif r < 0.5:
        return Operator.LOOP
    elif r < 0.75:
        return Operator.XOR
    else:
        return Operator.PARALLEL


def apply(parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> ProcessTree:
    """
    Generate a process tree

    Parameters
    ------------
    parameters
        Paramters of the algorithm, including:
            Parameters.REC_DEPTH -> current recursion depth
            Parameters.MIN_REC_DEPTH -> minimum recursion depth
            Parameters.MAX_REC_DEPTH -> maximum recursion depth
            Parameters.PROB_LEAF -> Probability to get a leaf

    Returns
    ------------
    tree
        Process tree
    """
    if parameters is None:
        parameters = {}

    rec_depth = exec_utils.get_param_value(Parameters.REC_DEPTH, parameters, 0)
    min_rec_depth = exec_utils.get_param_value(Parameters.MIN_REC_DEPTH, parameters, 1)
    max_rec_depth = exec_utils.get_param_value(Parameters.MAX_REC_DEPTH, parameters, 3)
    prob_leaf = exec_utils.get_param_value(Parameters.PROB_LEAF, parameters, 0.25)

    next_parameters = {Parameters.REC_DEPTH: rec_depth + 1, Parameters.MIN_REC_DEPTH: min_rec_depth,
                       Parameters.MAX_REC_DEPTH: max_rec_depth,
                       Parameters.PROB_LEAF: prob_leaf}

    is_leaf = False

    if min_rec_depth <= rec_depth <= max_rec_depth:
        r = random.random()
        if r < prob_leaf:
            is_leaf = True
    elif rec_depth > max_rec_depth:
        is_leaf = True

    if is_leaf:
        current_tree = ProcessTree(label=generate_random_string(6))
    elif rec_depth == 0:
        current_tree = ProcessTree(operator=Operator.SEQUENCE)
        start = ProcessTree(label=generate_random_string(6), parent=current_tree)
        current_tree.children.append(start)
        node = apply(parameters=next_parameters)
        node.parent = current_tree
        current_tree.children.append(node)
        end = ProcessTree(label=generate_random_string(6))
        end.parent = current_tree
        current_tree.children.append(end)
    else:
        o = get_random_operator()

        current_tree = ProcessTree(operator=o)
        if o == Operator.SEQUENCE:
            n_min = 2
            n_max = 6
            selected_n = random.randrange(n_min, n_max)
            for i in range(selected_n):
                child = apply(parameters=next_parameters)
                child.parent = current_tree
                current_tree.children.append(child)
        elif o == Operator.LOOP:
            do = apply(parameters=next_parameters)
            do.parent = current_tree
            current_tree.children.append(do)
            redo = apply(parameters=next_parameters)
            redo.parent = current_tree
            current_tree.children.append(redo)
            exit = ProcessTree(parent=current_tree)
            current_tree.children.append(exit)
        elif o == Operator.XOR:
            n_min = 2
            n_max = 5
            selected_n = random.randrange(n_min, n_max)
            for i in range(selected_n):
                child = apply(parameters=next_parameters)
                child.parent = current_tree
                current_tree.children.append(child)
        elif o == Operator.PARALLEL:
            n_min = 2
            n_max = 4
            selected_n = random.randrange(n_min, n_max)
            for i in range(selected_n):
                child = apply(parameters=next_parameters)
                child.parent = current_tree
                current_tree.children.append(child)
    return current_tree



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from enum import Enum
from pm4py.util import exec_utils
from pm4py.streaming.algo.conformance.footprints.variants import classic


class Variants(Enum):
    CLASSIC = classic


def apply(footprints, variant=Variants.CLASSIC, parameters=None):
    """
    Gets a footprints conformance checking object

    Parameters
    --------------
    footprints
        Footprints object (calculated from an entire log, from a process tree ...)
    variant
        Variant of the algorithm. Possible values: Variants.CLASSIC
    parameters
        Parameters of the algorithm

    Returns
    --------------
    fp_check_obj
        Footprints conformance checking object
    """
    return exec_utils.get_variant(variant).apply(footprints, parameters=parameters)
